@@ -1,9 +1,9 @@
 __author__ = 'Altertech, http://www.altertech.com/'
 __copyright__ = 'Copyright (C) 2018-2020 Altertech Group'
 __license__ = 'Apache License 2.0'
-__version__ = '2.0.28'
+__version__ = '2.0.32'
 
-from flask import request, jsonify, Response, abort
+from flask import request, jsonify, Response, abort, send_file, make_response
 from flask_restx import Api, Resource, reqparse
 
 import simplejson
@@ -14,7 +14,9 @@ flask.json = simplejson
 
 import base64
 import uuid
+import io
 import logging
+import datetime
 from functools import wraps
 
 from sqlalchemy import text as sql
@@ -35,7 +37,10 @@ from .core import subscription_get, subscription_list, subscription_create
 from .core import subscription_update, subscription_delete
 from .core import cleanup as core_cleanup, plugin_list
 
-from .core import json, is_parse_db_json, delete_everything
+from .core import json, is_parse_db_json, is_use_interval, delete_everything
+from .core import bucket_get, bucket_touch
+
+from .core import rq
 
 from functools import wraps
 
@@ -121,9 +126,9 @@ def _response_accepted():
 def public_method(f):
 
     @wraps(f)
-    def do(*args):
+    def do(*args, **kwargs):
         logger.debug(f'API call {get_real_ip()} {f.__qualname__}')
-        return f(*args, **(request.json if request.json else {}))
+        return f(*args, **kwargs)
 
     return do
 
@@ -182,6 +187,36 @@ def init():
             get_db()
             return '', 204
 
+    @ns_public.route('/file/<string:object_id>')
+    class File(Resource):
+
+        method_decorators = [public_method]
+
+        @api.response(200, 'serving file')
+        @api.response(404, 'file not found')
+        @api.produces(['*/*'])
+        def get(self, object_id):
+            try:
+                o = bucket_get(object_id, public=True)
+                buf = io.BytesIO()
+                buf.write(o['content'])
+                buf.seek(0)
+                bucket_touch(object_id)
+                expires = o['expires'].total_seconds() if is_use_interval(
+                ) else o['expires']
+                resp = make_response(
+                    send_file(buf,
+                              mimetype=o['mimetype'],
+                              cache_timeout=round(expires)))
+                for k, v in o['metadata'].items():
+                    resp.headers[f'x-roboger-{k}'] = str(v)
+                resp.headers['Content-Type'] = o['mimetype']
+                if o['fname']:
+                    resp.headers['roboger-file-name'] = o['fname']
+                return resp
+            except LookupError:
+                return 'no such file', 404
+
     @ns_public.route('/push')
     class Push(Resource):
 
@@ -229,7 +264,7 @@ def init():
         @api.response(429, 'recipient address is out of limit')
         @api.response(503, 'server error')
         @api.expect(p_push, validate=True)
-        def post(self, **kwargs):
+        def post(self):
             """
             Push event message
             """
@@ -280,26 +315,7 @@ def init():
                     return f'addr {a.addr} not found', 404
                 except OverlimitError as e:
                     return str(e), 429
-                for row in get_db().execute(sql("""
-                    SELECT plugin_name, config
-                    FROM subscription JOIN endpoint ON
-                        endpoint.id = subscription.endpoint_id JOIN addr ON
-                        endpoint.addr_id = addr.id WHERE
-                        addr.a=:a
-                        AND addr.active=1
-                        AND subscription.active = 1
-                        AND endpoint.active = 1
-                        AND (location=:location or location IS null)
-                        AND (tag=:tag or tag IS null)
-                        AND (sender=:sender or sender IS null)
-                        AND (
-                            (level=:level AND level_match='e') OR
-                            (level<:level and level_match='g') OR
-                            (level<=:level and level_match='ge') OR
-                            (level>:level and level_match='l') OR
-                            (level>=:level and level_match='le')
-                            )
-                            """),
+                for row in get_db().execute(sql(rq('push')),
                                             a=a.addr,
                                             location=a.location,
                                             tag=a.tag,
@@ -310,6 +326,7 @@ def init():
                          if is_parse_db_json() else row.config,
                          event_id=event_id,
                          addr=a.addr,
+                         addr_id=row.addr_id,
                          msg=a.msg,
                          subject=a.subject,
                          formatted_subject=formatted_subject,
