@@ -1,7 +1,7 @@
 __author__ = 'Altertech, http://www.altertech.com/'
 __copyright__ = 'Copyright (C) 2018-2020 Altertech Group'
 __license__ = 'Apache License 2.0'
-__version__ = '2.0.28'
+__version__ = '2.0.32'
 
 import platform
 import os
@@ -12,6 +12,8 @@ import threading
 import traceback
 import sqlalchemy
 import signal
+import datetime
+import time
 try:
     import rapidjson as json
 except:
@@ -26,7 +28,13 @@ from pathlib import Path
 from pyaltt2.crypto import gen_random_str
 from pyaltt2.network import parse_host_port, generate_netacl
 from pyaltt2.config import load_yaml, config_value, choose_file
+from pyaltt2.res import ResourceStorage
 from netaddr import IPNetwork
+from functools import partial
+from hashlib import sha256
+
+rs = ResourceStorage(mod='roboger')
+rq = partial(rs.get, resource_subdir='sql', ext='sql')
 
 SERVER_CONFIG_SCHEMA = {
     'type': 'object',
@@ -73,6 +81,15 @@ SERVER_CONFIG_SCHEMA = {
                     },
                     'additionalProperties': False,
                 },
+                'bucket': {
+                    'type': 'object',
+                    'properties': {
+                        'default-expires': {
+                            'type': 'integer'
+                        }
+                    },
+                    'additionalProperties': False
+                },
                 'secure-mode': {
                     'type': 'boolean'
                 },
@@ -137,6 +154,7 @@ logging.getLogger('urllib3').setLevel(logging.CRITICAL)
 _d = SimpleNamespace(db=None,
                      use_lastrowid=False,
                      parse_db_json=False,
+                     use_interval=False,
                      pool=None,
                      secure_mode=False,
                      ip_header=None,
@@ -154,6 +172,8 @@ dir_me = Path(__file__).absolute().parents[1].as_posix()
 
 db_lock = threading.RLock()
 
+default_bucket_expires = 86400
+
 logger = logging.getLogger('gunicorn.error')
 #logging.getLogger('roboger')
 
@@ -228,12 +248,21 @@ def load(fname=None):
             return False
 
     logger.info(f'CORE Roboger server {__version__} {product.build}')
-    fname = choose_file(
-        env='ROBOGER_CONFIG',
-        choices=[f'{dir_me}/etc/roboger.yml', '/usr/local/etc/roboger.yml'])
+    if not fname:
+        fname = choose_file(
+            env='ROBOGER_CONFIG',
+            choices=[f'{dir_me}/etc/roboger.yml', '/usr/local/etc/roboger.yml'])
     logger.debug(f'CORE using config file {fname}')
     server_config = load_yaml(fname, schema=SERVER_CONFIG_SCHEMA)
     config.update(server_config['roboger'])
+    config_value(config=config,
+                 config_path='/bucket/default-expires',
+                 in_place=True,
+                 default=default_bucket_expires)
+    config_value(env='ROBOGER_DB',
+                 config=config,
+                 config_path='/db',
+                 in_place=True)
     _d.secure_mode = config.get('secure-mode')
     _d.log_tracebacks = config.get('log-tracebacks')
     _d.limits = config.get('limits')
@@ -289,6 +318,8 @@ def load(fname=None):
         'sqlite') or config['db'].startswith('mysql')
     _d.parse_db_json = config['db'].startswith(
         'sqlite') or config['db'].startswith('mysql')
+    _d.use_interval = not config['db'].startswith(
+        'sqlite') and not config['db'].startswith('mysql')
     logger.debug(f'CORE database {_d.db}')
     get_db()
     thread_pool_size = config.get('thread-pool-size', default_thread_pool_size)
@@ -310,7 +341,7 @@ def spawn(*args, **kwargs):
 def init_db():
     from sqlalchemy import (Table, Column, BigInteger, Integer, Numeric, CHAR,
                             VARCHAR, MetaData, Float, ForeignKey, Index, JSON,
-                            Enum)
+                            Enum, DateTime, Interval, Boolean, LargeBinary)
     import enum
 
     class LevelMatch(enum.Enum):
@@ -320,6 +351,10 @@ def init_db():
         ge = 'ge'
         e = 'e'
 
+    if 'mysql' in _d.db.name:
+        from sqlalchemy.dialects.mysql import DATETIME, LONGBLOB
+        DateTime = partial(DATETIME, fsp=6)
+        LargeBinary = LONGBLOB
     meta = MetaData()
     addr = Table('addr',
                  meta,
@@ -389,6 +424,30 @@ def init_db():
                                 server_default='ge'),
                          mysql_engine='InnoDB',
                          mysql_charset='utf8mb4')
+    bucket = Table('bucket',
+                   meta,
+                   Column('id', CHAR(64), nullable=False, primary_key=True),
+                   Column('creator', VARCHAR(64), nullable=False),
+                   Column('addr_id',
+                          BigInteger().with_variant(Integer, 'sqlite'),
+                          ForeignKey('addr.id', ondelete='CASCADE')),
+                   Column('mimetype', VARCHAR(256), nullable=False),
+                   Column('fname', VARCHAR(256), nullable=True),
+                   Column('size',
+                          BigInteger().with_variant(Integer, 'sqlite'),
+                          nullable=False),
+                   Column('public', Boolean, nullable=False,
+                          server_default='0'),
+                   Column('metadata', JSON, nullable=True),
+                   Column('d', DateTime(timezone=True), nullable=False),
+                   Column('da', DateTime(timezone=True), nullable=True),
+                   Column('expires',
+                          Integer if _d.db.name in ('sqlite',
+                                                    'mysql') else Interval,
+                          nullable=True),
+                   Column('content', LargeBinary, nullable=False),
+                   mysql_engine='InnoDB',
+                   mysql_charset='utf8mb4')
     meta.create_all(_d.db)
 
 
@@ -428,6 +487,10 @@ def is_use_lastrowid():
     return _d.use_lastrowid
 
 
+def is_use_interval():
+    return _d.use_interval
+
+
 def is_parse_db_json():
     return _d.parse_db_json
 
@@ -484,7 +547,7 @@ def _safe_send(plugin_name, send_func, event_id, **kwargs):
 
 
 def check_addr_limit(addr, level, size):
-    a = addr['a']
+    a = addr['id']
     lim_c = addr['lim_c']
     lim_s = addr['lim_s']
     key_c = f'{a}.lim_c'
@@ -501,31 +564,36 @@ def check_addr_limit(addr, level, size):
     except Exception as e:
         logger.error(f'CORE Redis error: {e}')
         return
-    logger.debug(f'checking limits for {a}, current: '
+    logger.debug(f'checking limits for addr.id={a}, current: '
                  f'{current_c}, size: {current_s}, current msg level: {level}')
     try:
         if current_c >= lim_c:
-            logger.info(f'CORE address count overlimit {a} for all priorities')
+            logger.info(
+                f'CORE address count overlimit addr.id={a} for all priorities')
             raise OverlimitError(
-                f'Messages to {a} are limited by {lim_c} per '
+                f'Messages to {addr["a"]} are limited by {lim_c} per '
                 f'{_d.limits["period"]}. Limit has been reached')
         elif current_c >= lim_c - (lim_c / 100 *
                                    _d.limits['reserve']) and level < 30:
-            logger.info(f'CORE address count overlimit {a} for low-priority')
-            raise OverlimitError(f'Messages to {a} are limited by {lim_c} per '
-                                 f'{_d.limits["period"]}, '
-                                 f'{_d.limits["reserve"]}% are reserved for '
-                                 'WARNING and higher levels')
-        if current_s >= lim_s + size:
-            logger.info(f'CORE address size overlimit {a} for all priorities')
+            logger.info(
+                f'CORE address count overlimit addr.id={a} for low-priority')
             raise OverlimitError(
-                f'Messages to {a} are limited by {lim_s} bytes per '
+                f'Messages to {addr["a"]} are limited by {lim_c} per '
+                f'{_d.limits["period"]}, '
+                f'{_d.limits["reserve"]}% are reserved for '
+                'WARNING and higher levels')
+        if current_s >= lim_s + size:
+            logger.info(
+                f'CORE address size overlimit addr.id={a} for all priorities')
+            raise OverlimitError(
+                f'Messages to {addr["a"]} are limited by {lim_s} bytes per '
                 f'{_d.limits["period"]}. Limit has been reached')
         elif current_s >= lim_s - (lim_s / 100 *
                                    _d.limits['reserve']) and level < 30:
-            logger.info(f'CORE address size overlimit {a} for low-priority')
+            logger.info(
+                f'CORE address size overlimit addr.id={a} for low-priority')
             raise OverlimitError(
-                f'Messages to {a} are limited by {lim_s} bytes per '
+                f'Messages to {addr["a"]} are limited by {lim_s} bytes per '
                 f'{_d.limits["period"]}, '
                 f'{_d.limits["reserve"]}% are reserved for '
                 'WARNING and higher levels')
@@ -545,6 +613,7 @@ def reset_addr_limits():
 
 def cleanup():
     logger.debug('CORE cleanup')
+    bucket_cleanup()
     for k, v in plugins.items():
         safe_run_method(v, 'cleanup')
 
@@ -561,15 +630,12 @@ def db_list(*args, **kwargs):
 
 
 def delete_everything():
-    get_db().execute(sql("""
-            DELETE FROM addr
-            """))
+    get_db().execute(sql(rq('del')))
 
 
 def addr_get(addr_id=None, addr=None):
     result = get_db().execute(sql(
-        """SELECT id, a, active{} FROM addr WHERE id=:id or a=:a""".format(
-            ',lim_c,lim_s' if _d.limits else '')),
+        rq('addr.get').format(',lim_c,lim_s' if _d.limits else '')),
                               id=addr_id,
                               a=addr).fetchone()
     if result:
@@ -580,15 +646,13 @@ def addr_get(addr_id=None, addr=None):
 
 def addr_list():
     return db_list(
-        sql("""SELECT id, a, active{} FROM addr ORDER BY id""".format(
-            ',lim_c, lim_s' if _d.limits else '')))
+        sql(rq('addr.list').format(',lim_c, lim_s' if _d.limits else '')))
 
 
 def addr_create():
     addr = gen_random_str(64)
-    result = get_db().execute(sql("""
-            INSERT INTO addr (a) VALUES (:a) {}
-            """.format('' if is_use_lastrowid() else 'RETURNING id')),
+    result = get_db().execute(sql(
+        rq('addr.create').format('' if is_use_lastrowid() else 'RETURNING id')),
                               a=addr)
     i = result.lastrowid if is_use_lastrowid() else result.fetchone().id
     return i
@@ -598,9 +662,7 @@ def addr_change(addr_id=None, addr=None, to=None):
     if to is None:
         to = gen_random_str(64)
     try:
-        if get_db().execute(sql("""
-                UPDATE addr SET a=:new_a WHERE id=:id or a=:a
-                """),
+        if get_db().execute(sql(rq('addr.update.a')),
                             new_a=to,
                             id=addr_id,
                             a=addr).rowcount:
@@ -612,9 +674,7 @@ def addr_change(addr_id=None, addr=None, to=None):
 
 
 def addr_set_active(addr_id=None, addr=None, active=1):
-    if get_db().execute(sql("""
-            UPDATE addr SET active=:active WHERE id=:id or a=:a
-            """),
+    if get_db().execute(sql(rq('addr.update.active')),
                         active=active,
                         id=addr_id,
                         a=addr).rowcount:
@@ -625,29 +685,20 @@ def addr_set_active(addr_id=None, addr=None, active=1):
 
 def addr_set_limit(addr_id=None, addr=None, lim_c=None, lim_s=None):
     if lim_c is not None:
-        if not get_db().execute(sql("""
-                UPDATE addr SET lim_c=:lim_c WHERE id=:id or a=:a
-                """),
-                                lim_c=lim_c,
-                                id=addr_id,
-                                a=addr).rowcount:
+        if not get_db().execute(
+                sql(rq('addr.update.lim_c')), lim_c=lim_c, id=addr_id,
+                a=addr).rowcount:
             raise LookupError
     if lim_s is not None:
-        if not get_db().execute(sql("""
-                UPDATE addr SET lim_s=:lim_s WHERE id=:id or a=:a
-                """),
-                                lim_s=lim_s,
-                                id=addr_id,
-                                a=addr).rowcount:
+        if not get_db().execute(
+                sql(rq('addr.update.lim_s')), lim_s=lim_s, id=addr_id,
+                a=addr).rowcount:
             raise LookupError
     return addr_get(addr_id=addr_id, addr=addr)
 
 
 def addr_delete(addr_id=None, addr=None):
-    if not get_db().execute(sql("""
-            DELETE FROM addr WHERE id=:id or a=:a
-            """),
-                            id=addr_id,
+    if not get_db().execute(sql(rq('addr.delete')), id=addr_id,
                             a=addr).rowcount:
         raise LookupError
 
@@ -744,12 +795,10 @@ def endpoint_update(endpoint_id, data, validate_config=False, plugin_name=None):
 
 
 def endpoint_delete(endpoint_id):
-    if not get_db().execute(
-            sql("""
+    if not get_db().execute(sql("""
             DELETE FROM endpoint WHERE id=:id
             """),
-            id=endpoint_id,
-    ).rowcount:
+                            id=endpoint_id).rowcount:
         raise LookupError
 
 
@@ -862,10 +911,177 @@ def subscription_update(subscription_id, data):
 
 
 def subscription_delete(subscription_id):
-    if not get_db().execute(
-            sql("""
+    if not get_db().execute(sql("""
             DELETE FROM subscription WHERE id=:id
             """),
-            id=subscription_id,
-    ).rowcount:
+                            id=subscription_id).rowcount:
         raise LookupError
+
+
+def bucket_get(object_id, public=None):
+    """
+    Get bucket object
+
+    Args:
+        object_id: object ID
+        public: filter by "public" attr (None - don't filter)
+    Raises:
+        LookupError: if object is not found
+    """
+    if public is None:
+        xargs = {}
+        cond = ''
+    else:
+        xargs = {'public': public}
+        cond = """ AND public = :public"""
+    if _d.db.name == 'sqlite':
+        q = f"""SELECT
+            id, creator, mimetype, fname, size, public, metadata, d, da,
+            datetime(strftime('%s', d) + expires, 'unixepoch') as de,
+            strftime('%s', d) - :d + expires as expires, content
+            FROM bucket WHERE id=:object_id {cond} AND
+                DATETIME(STRFTIME('%s', d) + expires, 'unixepoch') >= :d"""
+        d = datetime.datetime.now()
+    elif _d.db.name == 'mysql':
+        q = f"""SELECT
+            id, creator, mimetype, fname, size, public, metadata, d, da,
+            DATE_FORMAT(FROM_UNIXTIME(UNIX_TIMESTAMP(d)),
+                '%Y-%m-%d %H:%i:%s.%f') as de,
+            UNIX_TIMESTAMP(d) - :d + expires as expires,
+                content
+            FROM bucket WHERE id=:object_id {cond} AND
+                UNIX_TIMESTAMP(d) + expires >= :d"""
+        d = time.time()
+        logger.info(q)
+        logger.info(d)
+    else:
+        q = f"""SELECT
+            id, creator, mimetype, fname, size, public, metadata, d, da,
+            d + expires as de, d - :d + expires as expires, content
+            FROM bucket WHERE id=:object_id {cond} AND d + expires >= :d"""
+        d = datetime.datetime.now()
+    result = get_db().execute(sql(q), object_id=object_id, d=d,
+                              **xargs).fetchone()
+    if result:
+        result = dict(result)
+        if is_parse_db_json():
+            result['metadata'] = json.loads(result['metadata'])
+        return result
+    else:
+        raise LookupError
+
+
+def bucket_put(content,
+               creator,
+               addr_id,
+               mimetype=None,
+               fname=None,
+               public=False,
+               metadata={},
+               expires=None):
+    """
+    Args:
+        content: file content (binary/text, required)
+        creator: object creator
+        addr_id: object addr id
+        mimetype: optional object mime type (auto-detected if not provided)
+        fname: optional object file name (served in "roboger-file-name" header)
+        public: if True, file is accessible with public HTTP API as
+            /file/{object_id}
+        metadata: elements, provided in this dict, are served in headers as
+            "x-roboger-{key}: {vaue}"
+        expires: object expiration (in seconds), if not provided, default
+            server expiration (by default: 86400 seconds) is set
+    Returns:
+        bucket object ID
+    """
+    if not isinstance(content, bytes) and not isinstance(content.bytearray):
+        content = str(content).encode()
+    h = sha256(content[:1024])
+    h.update(str(datetime.datetime.now()).encode())
+    h.update(str(time.perf_counter()).encode())
+    h.update(creator.encode())
+    h.update(str(addr_id).encode())
+    object_id = h.hexdigest()
+    if not mimetype:
+        import magic
+        mime = magic.Magic(mime=True)
+        mimetype = mime.from_buffer(content)
+        if not mimetype: mimetype = 'application/octet-stream'
+    if metadata is None:
+        metadata = {}
+    if is_use_interval():
+        expires = datetime.timedelta(
+            seconds=(expires if expires is not None else config['bucket']
+                     ['default-expires']))
+    else:
+        if expires is None:
+            expires = config['bucket']['default-expires']
+    get_db().execute(sql("""
+        INSERT INTO bucket
+                (id, creator, addr_id, mimetype, fname, size, public,
+                metadata, d, expires, content)
+            VALUES
+                (:object_id, :creator, :addr_id, :mimetype, :fname, :size,
+                :public, :metadata, :d, :expires, :content)
+        """),
+                     object_id=object_id,
+                     creator=creator,
+                     addr_id=addr_id,
+                     mimetype=mimetype,
+                     fname=fname,
+                     size=len(content),
+                     public=public,
+                     metadata=json.dumps(metadata),
+                     d=datetime.datetime.now(),
+                     expires=expires,
+                     content=content)
+    logger.debug(
+        f'CORE new bucket object {object_id} {mimetype} from {creator}')
+    return object_id
+
+
+def bucket_touch(object_id):
+    """
+    Mark object accessed (set access time)
+
+    Raises:
+        LookupError: if object is not found
+    """
+    if not get_db().execute(sql("""
+            UPDATE bucket SET da = :da WHERE id=:object_id
+            """),
+                            da=datetime.datetime.now(),
+                            object_id=object_id).rowcount:
+        raise LookupError
+
+
+def bucket_delete(object_id):
+    """
+    Delete object
+
+    Expired objects are also deleted automatically, when core cleaup is called.
+
+    Raises:
+        LookupError: if object is not found
+    """
+    if not get_db().execute(sql("""
+            DELETE FROM bucket WHERE id=:object_id
+            """),
+                            object_id=object_id).rowcount:
+        raise LookupError
+    logger.debug(f'CORE deleted bucket object {object_id}')
+
+
+def bucket_cleanup():
+    if _d.db.name == 'sqlite':
+        q = """DELETE FROM bucket WHERE
+                    DATETIME(STRFTIME('%s', d) + expires, 'unixepoch') < :d"""
+        d = datetime.datetime.now()
+    elif _d.db.name == 'mysql':
+        q = 'DELETE FROM bucket WHERE UNIX_TIMESTAMP(d) + expires < :d'
+        d = time.time()
+    else:
+        q = 'DELETE FROM bucket WHERE d + expires < :d'
+        d = datetime.datetime.now()
+    get_db().execute(sql(q), d=d)
